@@ -1,7 +1,7 @@
 /*
  * dname.h -- Domain name handling.
  *
- * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
@@ -73,8 +73,8 @@ const dname_type *dname_make_from_packet(region_type *region,
  * given buffer. Returns length in buffer or 0 on error.
  * buffer must be MAXDOMAINLEN+1 long.
  */
-int dname_make_wire_from_packet(uint8_t *buf, 
-				buffer_type *packet, 
+int dname_make_wire_from_packet(uint8_t *buf,
+				buffer_type *packet,
 				int allow_pointers);
 
 /*
@@ -92,8 +92,8 @@ const dname_type *dname_parse(region_type *region, const char *name);
 
 /*
  * parse ascii string to wireformat domain name (without compression ptrs)
- * returns 0 on failure, the length of the wireformat on success. 
- * the result is stored in the wirefmt which must be at least MAXDOMAINLEN 
+ * returns 0 on failure, the length of the wireformat on success.
+ * the result is stored in the wirefmt which must be at least MAXDOMAINLEN
  * in size. On failure, the wirefmt can be altered.
  */
 int dname_parse_wire(uint8_t* wirefmt, const char* name);
@@ -179,9 +179,9 @@ dname_label(const dname_type *dname, uint8_t label)
  * Return < 0 if LEFT < RIGHT, 0 if LEFT == RIGHT, and > 0 if LEFT >
  * RIGHT.  The comparison is case sensitive.
  *
- * Pre: left != NULL && right != NULL
+ * Pre: vleft != NULL && vright != NULL
  */
-int dname_compare(const dname_type *left, const dname_type *right);
+int dname_compare(const void *vleft, const void *right);
 
 
 /*
@@ -346,6 +346,21 @@ label_next(const uint8_t *label)
 const char *dname_to_string(const dname_type *dname,
 			    const dname_type *origin);
 
+/*
+ * Convert DNAME to its string representation.  This is a reentrant
+ * version of dname_to_string.  The buf argument is a pointer to a
+ * user defined result buffer capable of holding the string representation
+ * of a DNAME.  Due to escape sequences and such, this buffer is recommeneded
+ * to be at least 5 * MAXDOMAINLEN in size.
+ *
+ * If ORIGIN is provided and DNAME is a subdomain of ORIGIN the dname
+ * will be represented relative to ORIGIN.
+ *
+ * Pre: dname != NULL
+ */
+const char *dname_to_string_r(const dname_type *dname,
+			      const dname_type *origin,
+			      char *buf);
 
 /*
  * Create a dname containing the single label specified by STR
@@ -374,4 +389,13 @@ const dname_type *dname_replace(region_type* region,
 				const dname_type* src,
 				const dname_type* dest);
 
+#ifndef FULL_PREHASH
+/**
+ * Create a dname representing the wildcard form of the passed dname.
+ */
+int dname_make_wildcard(struct region *region,
+		    struct dname const *dname,
+		    struct dname const **wildcard);
+#endif
+
 #endif /* _DNAME_H_ */
