@@ -1,13 +1,13 @@
 /*
  * nsd.c -- nsd(8)
  *
- * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
  */
 
-#include <config.h>
+#include "config.h"
 
 #include <sys/types.h>
 #include <sys/param.h>
@@ -41,6 +41,7 @@
 #include <unistd.h>
 
 #include "nsd.h"
+#include "namedb.h"
 #include "options.h"
 #include "tsig.h"
 
@@ -106,7 +107,7 @@ version(void)
 	fprintf(stderr, "%s version %s\n", PACKAGE_NAME, PACKAGE_VERSION);
 	fprintf(stderr, "Written by NLnet Labs.\n\n");
 	fprintf(stderr,
-		"Copyright (C) 2001-2006 NLnet Labs.  This is free software.\n"
+		"Copyright (C) 2001-2011 NLnet Labs.  This is free software.\n"
 		"There is NO warranty; not even for MERCHANTABILITY or FITNESS\n"
 		"FOR A PARTICULAR PURPOSE.\n");
 	exit(0);
@@ -188,7 +189,7 @@ readpid(const char *file)
 		return -1;
 	}
 
-	pid = strtol(pidbuf, &t, 10);
+	pid = (pid_t) strtol(pidbuf, &t, 10);
 
 	if (*t && *t != '\n') {
 		return -1;
@@ -235,9 +236,18 @@ writepid(struct nsd *nsd)
 void
 unlinkpid(const char* file)
 {
-	if (file && unlink(file) == -1)
-		log_msg(LOG_ERR, "failed to unlink pidfile %s: %s",
-			file, strerror(errno));
+	int fd = -1;
+
+	if (file) {
+		/* truncate pidfile */
+		fd = open(file, O_WRONLY | O_TRUNC, 0644);
+		if (fd != -1)
+			close(fd);
+		/* unlink pidfile */
+		if (unlink(file) == -1)
+			log_msg(LOG_WARNING, "failed to unlink pidfile %s: %s",
+				file, strerror(errno));
+	}
 }
 
 /*
@@ -309,9 +319,67 @@ sig_handler(int sig)
  *
  */
 #ifdef BIND8_STATS
+
+#ifdef USE_ZONE_STATS
+static void
+fprintf_zone_stats(FILE* fd, zone_type* zone, time_t now)
+{
+	int i;
+
+	/* NSTATS */
+	fprintf(fd, "NSTATS %s %lu",
+		dname_to_string(domain_dname(zone->apex),0),
+		(unsigned long) now);
+
+	for (i = 0; i <= 255; i++) {
+		if (zone->st.qtype[i] != 0) {
+			fprintf(fd, " %s=%lu", rrtype_to_string(i),
+				zone->st.qtype[i]);
+		}
+	}
+	fprintf(fd, "\n");
+
+	/* XSTATS */
+	fprintf(fd, "XSTATS %s %lu"
+		" RR=%lu RNXD=%lu RFwdR=%lu RDupR=%lu RFail=%lu RFErr=%lu RErr=%lu RAXFR=%lu"
+		" RLame=%lu ROpts=%lu SSysQ=%lu SAns=%lu SFwdQ=%lu SDupQ=%lu SErr=%lu RQ=%lu"
+		" RIQ=%lu RFwdQ=%lu RDupQ=%lu RTCP=%lu SFwdR=%lu SFail=%lu SFErr=%lu SNaAns=%lu"
+		" SNXD=%lu RUQ=%lu RURQ=%lu RUXFR=%lu RUUpd=%lu\n",
+		dname_to_string(domain_dname(zone->apex),0),
+		(unsigned long) now,
+		zone->st.dropped,
+		(unsigned long)0, (unsigned long)0,
+		(unsigned long)0, (unsigned long)0,
+		(unsigned long)0, (unsigned long)0,
+		zone->st.raxfr,
+		(unsigned long)0, (unsigned long)0,
+		(unsigned long)0,
+		zone->st.qudp + zone->st.qudp6 - zone->st.dropped,
+		(unsigned long)0, (unsigned long)0,
+		zone->st.txerr,
+		zone->st.opcode[OPCODE_QUERY],
+		zone->st.opcode[OPCODE_IQUERY],
+		zone->st.wrongzone,
+		(unsigned long)0,
+		zone->st.ctcp + zone->st.ctcp6,
+		(unsigned long)0,
+		zone->st.rcode[RCODE_SERVFAIL],
+		zone->st.rcode[RCODE_FORMAT],
+		zone->st.nona,
+		zone->st.rcode[RCODE_NXDOMAIN],
+		(unsigned long)0, (unsigned long)0,
+		(unsigned long)0,
+		zone->st.opcode[OPCODE_UPDATE]);
+}
+#endif
+
 void
 bind8_stats (struct nsd *nsd)
 {
+#ifdef USE_ZONE_STATS
+	FILE* fd;
+	zone_type* zone;
+#endif
 	char buf[MAXSYSLOGMSGLEN];
 	char *msg, *t;
 	int i, len;
@@ -364,6 +432,23 @@ bind8_stats (struct nsd *nsd)
 			(unsigned long)0, (unsigned long)0, (unsigned long)0, nsd->st.opcode[OPCODE_UPDATE]);
 	}
 
+#ifdef USE_ZONE_STATS
+	/* ZSTATS */
+	log_msg(LOG_INFO, "ZSTATS %s", nsd->zonestatsfile);
+	if ((fd = fopen(nsd->zonestatsfile, "a")) ==  NULL ) {
+		log_msg(LOG_ERR, "cannot open zone statsfile %s: %s",
+			nsd->zonestatsfile, strerror(errno));
+		return;
+	}
+	/* Write stats per zone */
+	zone = nsd->db->zones;
+	while (zone) {
+		fprintf_zone_stats(fd, zone, now);
+		zone = zone->next;
+	}
+	fclose(fd);
+#endif
+
 }
 #endif /* BIND8_STATS */
 
@@ -384,8 +469,9 @@ main(int argc, char *argv[])
 #endif /* HAVE_GETPWNAM */
 
 	/* For initialising the address info structures */
-	struct addrinfo hints[MAX_INTERFACES];
-	const char *nodes[MAX_INTERFACES];
+	/* static so it can get very big without overflowing the stack */
+	static struct addrinfo hints[MAX_INTERFACES];
+	static const char *nodes[MAX_INTERFACES];
 	const char *udp_port = 0;
 	const char *tcp_port = 0;
 
@@ -401,6 +487,9 @@ main(int argc, char *argv[])
 	nsd.dbfile	= 0;
 	nsd.pidfile	= 0;
 	nsd.server_kind = NSD_SERVER_MAIN;
+#ifdef USE_ZONE_STATS
+	nsd.zonestatsfile = 0;
+#endif
 
 	for (i = 0; i < MAX_INTERFACES; i++) {
 		memset(&hints[i], 0, sizeof(hints[i]));
@@ -490,7 +579,7 @@ main(int argc, char *argv[])
 			if (hex_pton(optarg, nsd.nsid, nsd.nsid_len) == -1) {
 				error("hex string cannot be parsed '%s' in NSID.", optarg);
 			}
-                       break;
+			break;
 		case 'l':
 			nsd.log_filename = optarg;
 			break;
@@ -657,6 +746,11 @@ main(int argc, char *argv[])
 	if(nsd.st.period == 0) {
 		nsd.st.period = nsd.options->statistics;
 	}
+#ifdef USE_ZONE_STATS
+	if (nsd.zonestatsfile == 0) {
+		nsd.zonestatsfile = nsd.options->zonestatsfile;
+	}
+#endif /* USE_ZONE_STATS */
 #endif /* BIND8_STATS */
 #ifdef HAVE_CHROOT
 	if(nsd.chrootdir == 0) nsd.chrootdir = nsd.options->chroot;
@@ -863,8 +957,11 @@ main(int argc, char *argv[])
 	log_open(LOG_PID, FACILITY, nsd.log_filename);
 	if (!nsd.log_filename)
 		log_set_log_function(log_syslog);
-	else if (nsd.uid && nsd.gid)
-		(void) chown(nsd.log_filename, nsd.uid, nsd.gid);
+	else if (nsd.uid && nsd.gid) {
+		if(chown(nsd.log_filename, nsd.uid, nsd.gid) != 0)
+			VERBOSITY(2, (LOG_WARNING, "chown %s failed: %s",
+				nsd.log_filename, strerror(errno)));
+	}
 
 	/* Do we have a running nsd? */
 	if ((oldpid = readpid(nsd.pidfile)) == -1) {
@@ -950,8 +1047,23 @@ main(int argc, char *argv[])
 		exit(1);
 	}
 
-#ifdef HAVE_CHROOT
+	/* Set user context */
+#ifdef HAVE_GETPWNAM
+	if (*nsd.username) {
+#ifdef HAVE_SETUSERCONTEXT
+		/* setusercontext does initgroups, setuid, setgid, and
+		 * also resource limits from login config, but we
+		 * still call setresuid, setresgid to be sure to set all uid */
+		if (setusercontext(NULL, pwd, nsd.uid,
+			LOGIN_SETALL & ~LOGIN_SETUSER & ~LOGIN_SETGROUP) != 0)
+			log_msg(LOG_WARNING, "unable to setusercontext %s: %s",
+				nsd.username, strerror(errno));
+#endif /* HAVE_SETUSERCONTEXT */
+	}
+#endif /* HAVE_GETPWNAM */
+
 	/* Chroot */
+#ifdef HAVE_CHROOT
 	if (nsd.chrootdir && strlen(nsd.chrootdir)) {
 		int l = strlen(nsd.chrootdir);
 		int ret = 0;
@@ -971,10 +1083,18 @@ main(int argc, char *argv[])
 		nsd.options->xfrdfile += l;
 		nsd.options->difffile += l;
 
+#ifdef HAVE_TZSET
+		/* set timezone whilst not yet in chroot */
+		tzset();
+#endif
 		if (chroot(nsd.chrootdir)) {
 			log_msg(LOG_ERR, "unable to chroot: %s", strerror(errno));
 			exit(1);
 		}
+		if (chdir("/")) {
+			log_msg(LOG_ERR, "unable to chdir to chroot: %s", strerror(errno));
+			exit(1);
+		}
 		DEBUG(DEBUG_IPC,1, (LOG_INFO, "changed root directory to %s",
 			nsd.chrootdir));
 	}
@@ -1001,20 +1121,11 @@ main(int argc, char *argv[])
 	/* Drop the permissions */
 #ifdef HAVE_GETPWNAM
 	if (*nsd.username) {
-#ifdef HAVE_SETUSERCONTEXT
-		/* setusercontext does initgroups, setuid, setgid, and
-		 * also resource limits from login config, but we
-		 * still call setresuid, setresgid to be sure to set all uid */
-		if (setusercontext(NULL, pwd, nsd.uid, LOGIN_SETALL) != 0)
-			log_msg(LOG_WARNING, "unable to setusercontext %s: %s",
-				nsd.username, strerror(errno));
-#else /* !HAVE_SETUSERCONTEXT */
- #ifdef HAVE_INITGROUPS
+#ifdef HAVE_INITGROUPS
 		if(initgroups(nsd.username, nsd.gid) != 0)
 			log_msg(LOG_WARNING, "unable to initgroups %s: %s",
 				nsd.username, strerror(errno));
- #endif /* HAVE_INITGROUPS */
-#endif /* HAVE_SETUSERCONTEXT */
+#endif /* HAVE_INITGROUPS */
 		endpwent();
 
 #ifdef HAVE_SETRESGID
@@ -1039,7 +1150,7 @@ main(int argc, char *argv[])
 
 		DEBUG(DEBUG_IPC,1, (LOG_INFO, "dropped user privileges, run as %s",
 			nsd.username));
-    }
+	}
 #endif /* HAVE_GETPWNAM */
 
 	if (server_prepare(&nsd) != 0) {
