@@ -1,7 +1,7 @@
 /*
  * options.h -- nsd.conf options definitions and prototypes
  *
- * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
@@ -10,7 +10,7 @@
 #ifndef OPTIONS_H
 #define OPTIONS_H
 
-#include <config.h>
+#include "config.h"
 #include <stdarg.h>
 #include "region-allocator.h"
 #include "rbtree.h"
@@ -48,14 +48,20 @@ struct nsd_options {
 	const char* logfile;
 	int server_count;
 	int tcp_count;
+	int tcp_query_count;
+	int tcp_timeout;
+	size_t ipv4_edns_size;
+	size_t ipv6_edns_size;
 	const char* pidfile;
 	const char* port;
 	int statistics;
+	const char* zonestatsfile;
 	const char* chroot;
 	const char* username;
 	const char* zonesdir;
 	const char* difffile;
 	const char* xfrdfile;
+	const char* nsid;
 	int xfrd_reload_timeout;
 
 	region_type* region;
@@ -82,6 +88,7 @@ struct zone_options {
 	acl_options_t* provide_xfr;
 	acl_options_t* outgoing_interface;
 	uint8_t allow_axfr_fallback;
+	uint8_t notify_retry;
 };
 
 union acl_addr_storage {
@@ -132,9 +139,7 @@ struct key_options {
 	const char* name;
 	const char* algorithm;
 	const char* secret;
-#ifdef TSIG
 	struct tsig_key* tsig_key;
-#endif
 };
 
 /*
@@ -172,14 +177,18 @@ zone_options_t* zone_options_create(region_type* region);
 zone_options_t* zone_options_find(nsd_options_t* opt, const struct dname* apex);
 key_options_t* key_options_create(region_type* region);
 key_options_t* key_options_find(nsd_options_t* opt, const char* name);
+
+#if defined(HAVE_SSL)
 /* tsig must be inited, adds all keys in options to tsig. */
 void key_options_tsig_add(nsd_options_t* opt);
+#endif
 
 /* check acl list, acl number that matches if passed(0..),
  * or failure (-1) if dropped */
 /* the reason why (the acl) is returned too (or NULL) */
 int acl_check_incoming(acl_options_t* acl, struct query* q,
 	acl_options_t** reason);
+int acl_addr_matches_host(acl_options_t* acl, acl_options_t* host);
 int acl_addr_matches(acl_options_t* acl, struct query* q);
 int acl_key_matches(acl_options_t* acl, struct query* q);
 int acl_addr_match_mask(uint32_t* a, uint32_t* b, uint32_t* mask, size_t sz);
