@@ -8,7 +8,7 @@
  */
 
 
-#include <config.h>
+#include "config.h"
 #include <stdlib.h>
 #include <ctype.h>
 
@@ -16,7 +16,6 @@
 #include "tsig-openssl.h"
 #include "dns.h"
 #include "packet.h"
-#include "query.h"
 
 static region_type *tsig_region;
 
