@@ -37,7 +37,6 @@
 #include "rdata.h"
 #include "region-allocator.h"
 #include "util.h"
-#include "zparser.h"
 #include "options.h"
 #include "nsec3.h"
 
@@ -391,13 +390,13 @@ zparser_conv_text(region_type *region, const char *text, size_t len)
 	if (len > 255) {
 		zc_error_prev_line("text string is longer than 255 characters,"
 				   " try splitting it into multiple parts");
-	} else {
-		uint8_t *p;
-		r = alloc_rdata(region, len + 1);
-		p = (uint8_t *) (r + 1);
-		*p = len;
-		memcpy(p + 1, text, len);
+		len = 255;
 	}
+	uint8_t *p;
+	r = alloc_rdata(region, len + 1);
+	p = (uint8_t *) (r + 1);
+	*p = len;
+	memcpy(p + 1, text, len);
 	return r;
 }
 
@@ -688,7 +687,7 @@ zparser_conv_loc(region_type *region, char *str)
 				zc_error_prev_line("space expected after seconds");
 				return NULL;
 			}
-
+			/* No need for precision specifiers, it's a double */
 			if (sscanf(start, "%lf", &d) != 1) {
 				zc_error_prev_line("error parsing seconds");
 			}
